#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dogma import (
    GeneticCode,
    DEFAULT_GENETIC_CODE,
    Codon,
)


class AminoAcid:
    """
    Amino acids are the building blocks of proteins.

    Attributes
    ++++++++++
    label - str - One letter string label denoting acid acid identity
    members - list - List of
    Methods
    +++++++

    """

    def __init__(self,
                 data=None,
                 genetic_code=None,
                 triplet_string_is_dna=True):

        if not isinstance(genetic_code, GeneticCode):
            genetic_code = DEFAULT_GENETIC_CODE
        self.genetic_code = genetic_code

        # input data is string of length 1, an amino acid letter
        # AminoAcid('A')
        if isinstance(data, str) and len(data) == 1:

            self.label = data
            self.members = [data]
            self.proportions = [1]
            self.codon_labels = self.get_synonymous_codons()
            self.codons = [self.get_synonymous_codons('Codon')]
            self.composition = dict(zip(self.members, self.proportions))

        # input data is a string of length 3, a codon label
        # AminoAcid('GGG')
        elif (isinstance(data, str) and
              len(data) == 3 and
              triplet_string_is_dna):

            self.codons = [Codon(data, genetic_code)]
            self.codon_labels = self.codons[0].members
            self.composition = self.codons[0].translate()
            self.members = sorted(self.composition.keys())
            self.proportions = [self.composition[_] for _ in self.members]
            self.label = genetic_code[data]

        # input data is a Codon object
        # AminoAcid(Codon('GGG'))
        elif isinstance(data, Codon):

            self.codons = [data]
            self.codon_labels = self.codons[0].members
            self.composition = self.codons[0].translate()
            self.members = sorted(self.composition.keys())
            self.proportions = [self.composition[_] for _ in self.members]
            self.label = genetic_code[data.label]

        self.degenerate = self.is_degenerate()
        self.equimolar = self.is_equimolar()

    def get_synonymous_codons(self, output_type='string'):
        """
        Returns list of triplet strings or Codon instances.

        Each synonymous codon encodes a nonzero amino acid.
        """
        codon_strings = []
        for member in self.members:
            codon_strings += self.genetic_code.code_reversed[member]

        if output_type == 'string':
            return codon_strings
        elif output_type == 'Codon':
            return Codon(codon_strings, self.genetic_code)
        elif output_type == 'Codons':
            return [Codon(_, self.genetic_code) for _ in codon_strings]
        return None

    def is_degenerate(self):
        """
        Amino acid is degenerate if there are multiple amino acid members
        have nonzero composition values.
        """
        return len(self.members)

    def is_equimolar(self):
        """
        Amino acid is equimolar if all nonzero amino acid members
        have equal abundance.
        """
        return max(self.proportions) == min(self.proportions)

    def __str__(self):

        print(self.label)
        print(self.composition)

    def __repr__(self):
        l, c = self.label, self.composition
        return f'AminoAcid(label="{l}", composition={c})'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from itertools import product
from collections import defaultdict
import decimal
from random import choices

from dogma import (
    DEFAULT_DECIMAL_PRECISION,
    GeneticCode,
    DEFAULT_GENETIC_CODE,
    Nucleotide,
    DEGENERATE_NUCLEOTIDE_CODE,
    DEGENERATE_NUCLEOTIDE_CODE_REVERSED,
    is_valid_nucleotide_string,
    combine_nucleotides,
    STANDARD_CODONS,
    DEFAULT_CODON_LABEL
)

decimal.getcontext().prec = DEFAULT_DECIMAL_PRECISION


class Codon:
    """
    A sequence of three nucleotides.
    """

    def __init__(self, data=None, genetic_code=None):
        """
        Multiple input formats are acceptable
            n = Nucleotide('N')
            Codon([n,n,n]) == Codon((n, n, n))
        """

        if not isinstance(genetic_code, GeneticCode):
            genetic_code = DEFAULT_GENETIC_CODE
        self.genetic_code = genetic_code

        # aaa = Codon('AAA')
        # naa = Codon('NAA')
        # Xaa = combine_codons(aaa, naa)  # is dict
        # Codon(Xaa)
        if (isinstance(data, dict) and
                all([_ in STANDARD_CODONS for _ in data])):

            self.bases = None
            self.label = DEFAULT_CODON_LABEL
            self.members = list(data.keys())
            self.proportions = [data[_] for _ in self.members]

        # Codon(['AAA', 'GGG'])
        elif (isinstance(data, list) and
                all([isinstance(_, str) for _ in data]) and
                all([len(_) == 3 for _ in data]) and
                all([_ in STANDARD_CODONS for _ in data])):

            self.bases = None
            self.label = DEFAULT_CODON_LABEL
            self.members = list(set(data))
            self.proportions = [data.count(_) for _ in self.members]

        # Codon('AAA') or Codon('NNK') or ...
        # Codon([Nucleotide('A'),]*3) or ...
        # data as iterable of length 3 , or Nuc(),
        # or parameters that generates valid Nuc()
        elif (isinstance(data, (str, list, tuple)) and
                len(data) == 3):

            self.bases = [_
                          if isinstance(_, Nucleotide)
                          else Nucleotide(_)
                          for _ in data]
            self.label = ''.join([_.label for _ in self.bases])
            self.members = self.get_members()
            self.proportions = self.get_proportions()

        # Codon() --> Codon([Nucleotide(),]*3) or ...
        else:
            raise 'Error, Codon() parameters not valid'

        self.composition = dict(zip(self.members, self.proportions))

        self.degenerate = self.is_degenerate()
        self.equimolar = self.is_equimolar()

    def get_members(self):
        """
        Returns a list of codons (as triplet strings) with nonzero proportion.
        """
        b1, b2, b3 = [b.get_members() for b in self.bases]
        return [''.join(_) for _ in product(b1, b2, b3)]

    def get_proportions(self):
        """
        Returns a list of codon proportions, aligned with self.members
        """
        data = []
        for member in self.members:
            proportion = 1
            for m, b in zip(member, self.bases):
                proportion *= b.composition[m]
            data.append(proportion)
        return data

    def translate(self, dtype=None):
        """
        Returns a dictionary mapping amino acids (as single-letter codes)
        with proportions.
        """
        data = defaultdict(float)
        for c, p in zip(self.members, self.proportions):
            a = self.genetic_code[c]
            data[a] += p

        if dtype == 'decimal':
            data = {k: decimal.Decimal(v) for k, v in data.items()}
        elif dtype == 'float':
            data = {k: float(v) for k, v in data.items()}
        elif dtype == 'int':
            data = {k: int(v) for k, v in data.items()}
        elif dtype == 'object':
            pass
        elif dtype is None:
            data = {k: v for k, v in data.items()}
        else:
            # print('error', dtype, data)
            pass

        return data

    def samples(self, k=1):
        return [self.sample() for _ in range(k)]

    def sample(self):
        """
        Returns a single non-degenerate oligonucleotide string.
        """
        return choices(self.members, self.proportions)[0]

    def is_degenerate(self):
        """
        Codon is degenerate if there are multiple nonzero standard codons
        in composition.
        """
        return len(self.members) > 1

    def is_equimolar(self):
        """
        Codon is equimolar if all nonzero standard codons in composition have
        equal abundance.
        """
        return max(self.proportions) == min(self.proportions)

    def copy(self):
        return Codon([b.copy() for b in self.bases], self.genetic_code)

    def __str__(self):
        b0, b1, b2 = self.bases
        # return f'Codon(\n\t{b0}\n\t{b1}\n\t{b2}\n)'
        return f'Codon(\n\t{b0}\n\t{b1}\n\t{b2})'


def combine_codons(*codons, proportions=None):
    """
    codons unpacked to standard codons and proportions added

    ex: AAA + CCC + GGG + TTT --> [AAA, CCC, GGG, TTT]
    ex: NNN + AAA --> {AAA: 2 , AAC:1, AAG:1,.... TTT:1}
    """
    genetic_code = codons[0].genetic_code
    if proportions is None:
        proportions = [1, ] * len(codons)

    data = defaultdict(float)
    for codon, p in zip(codons, proportions):
        for sub_codon, sub_proportion in codon.composition.items():
            data[sub_codon] += p * sub_proportion

    return Codon(data, genetic_code)


def nucleotide_string_to_codons(s, output='Codon'):
    if is_valid_nucleotide_string(s):
        codon_strings = [s[3 * i:3 * i + 3] for i in range(len(s) // 3)]

        if output == 'Codon':
            return [Codon(*_) for _ in codon_strings]
        elif output == 'strings':
            return codon_strings


def degenerate_codon_string_to_standard_members(c):
    """
    Converts degenerate codon string to list of standard codons.
    """
    b1, b2, b3 = [DEGENERATE_NUCLEOTIDE_CODE[b] for b in c]
    return [''.join(_) for _ in product(b1, b2, b3)]


def combine_codon_labels(c1, c2):
    """
    Combines codon labels, assuming bases mixed randomly
    Ex: AAA + CCC + GGG + TTT -> NNN = (AAA, AAC, ....)
    Ex: ABC + CCC --> MBC
    """
    letters = []
    for b1, b2 in zip(c1, c2):
        b1 = b1.upper()
        b2 = b2.upper()
        _ = DEGENERATE_NUCLEOTIDE_CODE[b1] + DEGENERATE_NUCLEOTIDE_CODE[b2]
        base_letters = ''.join(sorted(list(set(_))))
        letters.append(DEGENERATE_NUCLEOTIDE_CODE_REVERSED[base_letters])
    return ''.join(letters)


def merge_codons(c1, c2, proportions=[1, 1], genetic_code=None):
    if genetic_code is None:
        genetic_code = DEFAULT_GENETIC_CODE
    nucs = [combine_nucleotides(n1, n2, proportions=proportions)
            for n1, n2 in zip(c1.bases, c2.bases)]
    return Codon(nucs, genetic_code)



import pandas as pd
from dogma import (
    Codon,
    GeneticCode,
    DEFAULT_GENETIC_CODE,
    NCBI_CODONS,
    DEGENERATE_CODONS
)


class CodonOption(Codon):
    def __init__(self, data=None, genetic_code=None):
        super().__init__(data, genetic_code)

        self.amino_acid_profile = self.translate()
        self.aa_string = self.get_aa_string()
        self.aas_string = self.get_aas_string()
        self.aa_profile_string = self.get_aa_profile_string()

    def get_aa_string(self):
        aas = sorted([_ for _, v in self.amino_acid_profile.items() if v > 0])
        return ''.join(aas)

    def get_aas_string(self):
        aas = [self.genetic_code[_] for _ in self.members]
        return ''.join(sorted(aas))

    def get_aa_profile_string(self):
        aas = sorted([_ for _, v in self.amino_acid_profile.items() if v > 0])
        output = [f'{aa}:{self.amino_acid_profile[aa]},' for aa in aas]
        return ''.join(output)

    def is_aa_equimolar(self):
        vals = [_ for _ in self.amino_acid_profile.values() if _ > 0]
        return max(vals) == min(vals)


class CodonSelector:
    """
    Container for the selection of degenerate codons within a genetic code that
    fulfill filtering criteria, such as exlcuding stop codons, rare codons, or
    must include specific amino acids.
    """

    def __init__(self,
                 genetic_code=DEFAULT_GENETIC_CODE,
                 must_include=[],
                 must_exclude=[],
                 scope=DEGENERATE_CODONS):

        self.genetic_code = genetic_code
        self.amino_acids = genetic_code.amino_acids
        self.codons = genetic_code.codons

        self.must_exclude = must_exclude
        self.must_include = must_include

        self.scope = scope
        self.table = self.make_table()
        self.filter()

    def parse_criteria(self):
        self.excluded_amino_acids = []
        self.excluded_codons = []
        for _ in self.must_exclude:
            if _ in self.amino_acids:
                self.excluded_amino_acids.append(_)
            elif _ in self.codons:
                self.excluded_codons.append(_)

        self.included_amino_acids = []
        self.included_codons = []
        for _ in self.must_include:
            if _ in self.amino_acids:
                self.included_amino_acids.append(_)
            elif _ in self.codons:
                self.included_codons.append(_)

    def make_table(self, dcs=None):
        if dcs is None:
            dcs = [CodonOption(_, self.genetic_code) for _ in self.scope]

        data = {
            'Labels': [_.label for _ in dcs],
            'AAs': [_.get_aas_string() for _ in dcs],
            'uAAs': [_.aa_string for _ in dcs]
            }
        data['nCodons'] = [len(_) for _ in data['AAs']]
        data['nuAAs'] = [len(_) for _ in data['uAAs']]
        data['Stops'] = ['*' in _ for _ in data['uAAs']]
        data['Equimolar'] = [_.is_aa_equimolar() for _ in dcs]

        return pd.DataFrame(data)

    def filter(self):
        self.parse_criteria()

        self.filtered_options = []
        for dc in self.scope:
            c = CodonOption(dc, self.genetic_code)
            data = c.translate()
            PASSES_FILTERS = True

            for _ in self.excluded_codons:
                if _ in c.members:
                    PASSES_FILTERS = False
                    break

            if PASSES_FILTERS:
                for _ in self.excluded_amino_acids:
                    if data.get(_, 0) > 0:
                        PASSES_FILTERS = False
                        break

            if PASSES_FILTERS:
                for _ in self.included_codons:
                    if _ not in c.members:
                        PASSES_FILTERS = False
                        break

            if PASSES_FILTERS:
                for _ in self.included_amino_acids:
                    if data.get(_, 0) == 0:
                        PASSES_FILTERS = False
                        break

            if PASSES_FILTERS:
                self.filtered_options.append(c)

        self.filtered_table = self.make_table(self.filtered_options)


def main():
    CS = CodonSelector(must_exclude=list('*'))
    CS.filter()
    return CS


if __name__ == '__main__':
    cs = main()



from functools import reduce
from random import choices

from dogma import (
    GeneticCode,
    DEFAULT_GENETIC_CODE,
    Oligonucleotide,
    Codon,
    AminoAcid,
)


class Protein:
    """
    A sequence of AminoAcid objects.

    N-terminal --> C-terminal list

    Parameters
    ----------
    data:
    genetic_code: GeneticCode object
    data_is_dna: flag for processing data

    Attributes
    ---------
    oligonucleotide: (opt) Oligonucleotide object.
                     If not exactly determined, consist of all possible dna
    """

    def __init__(self, data, genetic_code=None, data_is_dna=False):

        if not isinstance(genetic_code, GeneticCode):
            genetic_code = DEFAULT_GENETIC_CODE
        self.genetic_code = genetic_code

        if all([isinstance(_, str) for _ in data]):

            # Protein('NNK', data_is_dna=True)
            if data_is_dna:
                self.oligonucleotide = Oligonucleotide(data, genetic_code)
                self.codons = oligonucleotide_to_codons(self.oligonucleotide)
                self.residues = [AminoAcid(c.translate()) for c in self.codons]

            # Protein('NNK')
            else:
                self.residues = [AminoAcid(_, genetic_code) for _ in data]
                self.codons = [a.get_synonymous_codons(Codon)
                               for a in self.residues]
                self.oligonucleotide = None

        # Protein(Oligonucleotide('NNK'))
        elif isinstance(data, Oligonucleotide):
            self.oligonucleotide = data
            self.codons = oligonucleotide_to_codons(self.oligonucleotide)
            self.residues = [AminoAcid(c.translate()) for c in self.codons]

        # Protein(AminoAcid('A'))
        elif isinstance(data, AminoAcid):
            self.residues = [data]
            self.oligonucleotide = None
            self.codons = oligonucleotide_to_codons(self.oligonucleotide)

        else:
            self.residues = []
            self.oligonucleotide = None
            self.codons = []

        self.label = ''.join([_.label for _ in self.residues])
        self.length = len(self.residues)
        self.degenerate = self.is_degenerate()

    def is_degenerate(self):
        return any([a.is_degenerate() for a in self.residues])

    def get_synonymous_codons(self):
        return [codon for codon in [a.get_synonymous_codons()
                for a in self.amino_acids]]

    def samples(self, k=1):
        return [self.sample() for _ in range(k)]

    def sample(self):
        return choices(self.letters, self.proportions)

    def __str__(self):
        return self.label

    def __repr__(self):
        return f'Protein(label={self.label}, composition={self.composition})'


def calculate_protein_degeneracy(protein, oligonucleotide=None):
    """
    Calculates degeneracy of protein sequences for a specified
    oligonucleotide design.

    If no oligonucleotide is specified, defaults to a fully randomized
    ('NNN') scheme with a length 3 times as large as the protein, and
    the default genetic code is used.
    """

    if isinstance(protein, Protein):
        protein = protein.label  # how to handle '_'?

    if oligonucleotide is None:
        oligonucleotide = Oligonucleotide('NNN' * len(protein))

    data = oligonucleotide.amino_acid_profile

    return reduce(lambda x, y: x * y, [aa_composition[aa]
                  for aa_composition, aa in zip(data, protein)],)


def oligonucleotide_to_codons(oligo):
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import choices

from dogma import (
    rescale,
    STANDARD_NUCLEOTIDES,
    DEFAULT_NUCLEOTIDE_LABEL,
    DEGENERATE_NUCLEOTIDES,
    DEGENERATE_NUCLEOTIDE_CODE_COMPOSITION,
    DEGENERATE_NUCLEOTIDE_CODE_REVERSED
)


class Nucleotide:
    """
    Nucleotides are the building blocks of oligonucleotides.

    Parameters
    ----------
    data: multiple inputs available, see self._process_input()

    Attributes
    ----------
    label: str, len=1
        single letter code for nucleotide
    composition: dictionary
        keys are single letters refering to undegenerate nucleotides
        values are numerical values related to abundance proportion
    members
        sorted list of nonzero nucleotides (as single letter codes)
    proportions
        list of member proportions, aligned with members
    degenerate: Boolean
        whether multiple standard nucleotides are nonzero
    equimolar: Boolean
        whether all nonzero standard nucleotide members have equal proportions

    Usage
    -----
    a = Nucleotide('A')
        a.label --> 'A'
        a.composition --> {'A':1}
        a.members --> ['A']
        a.proportions --> [1]
        a.degenerate --> False
        a.equimolar --> True
    a = Nucleotide({'A':1})

    n = Nucleotide('N')
        a.label --> 'N'
        a.composition --> {'A':1, 'C':1, 'G':1, 'T':1}
        a.members --> ['A', 'C', 'G', 'T']
        a.proportions --> [1, 1, 1, 1]
        a.degenerate --> True
        a.equimolar --> True

    n = Nucleotide({'A':0.1, 'C':0.5, 'G':0.4})
        a.label --> 'v'
        a.composition --> {'A':0.1, 'C':0.5, 'G':0.4}
        a.members --> ['A', 'C', 'G', 'T']
        a.proportions --> [0.1, 0.5, 0.4]
        a.degenerate --> True
        a.equimolar --> False
    """

    def __init__(self, data=None):

        # define self.label and self.composition by processing
        # data input parameter
        self._process_input(data)

        self.members = self.get_members()
        self.proportions = self.get_proportions()

        self.degenerate = self.is_degenerate()
        self.equimolar = self.is_equimolar()

    def _process_input(self, data):
        """
        Various input data formats are supported:
            Nucleotide('A')
            Nucleotide({'A': 3, 'C': 1})
        """

        # ex: Nucleotide('A')
        if isinstance(data, str) and len(data) == 1:

            label = data.upper().replace('U', 'T')
            composition = DEGENERATE_NUCLEOTIDE_CODE_COMPOSITION[label]

        # ex: Nucleotide({'A':3, 'C':1})
        elif (isinstance(data, dict) and
                is_valid_nucleotide_string(''.join(data.keys())) and
                all([isinstance(_, (int, float)) for _ in data.values()])):

            composition = data
            label = nucleotide_composition_to_letter(composition)

        else:
            label = DEFAULT_NUCLEOTIDE_LABEL
            composition = DEGENERATE_NUCLEOTIDE_CODE_COMPOSITION[label]

        self.label = label
        self.composition = composition

    def get_members(self):
        """
        Returns a sorted string of nucleotide members that are have nonzero
        frequency within nucleotide.
        """
        nonzero_members = [k for k, v in self.composition.items() if v > 0]
        return ''.join(sorted(nonzero_members))

    def get_proportions(self):
        """
        Returns a list of proportions aligned with self.members
        """
        return [self.composition[_] for _ in self.members]

    def is_degenerate(self):
        """
        Nucleotide is degenerate if multiple standard nucleotides
        have nonzero proportions.
        """
        return len(self.members) > 1

    def is_equimolar(self):
        """
        Nucleotide is equimolar if the proportion of each nonzero nucleotide
        values are equal.
        """
        nonzero_values = [_ for _ in self.composition.values() if _ > 0]
        return max(nonzero_values) == min(nonzero_values)

    def samples(self, k=1):
        """
        Randomly selects `k` members based on Nucleotide composition.
        """
        return [self.sample() for _ in range(k)]

    def sample(self):
        """
        Randomly selects a members based on Nucleotide composition.
        """
        return choices(self.members, self.proportions)[0]

    def copy(self):
        return Nucleotide(self.composition)

    def __str__(self):
        return f'Nucleotide(label={self.label}, '\
               f'composition={self.composition})'

    def __repr__(self):
        return f'Nucleotide(label={self.label}, '\
               f'composition={self.composition})'


def is_nondegenerate_nucleotide_string(n):
    n = n.upper().replace('U', 'T')
    return all([_ in STANDARD_NUCLEOTIDES for _ in n])


def is_degenerate_nucleotide_string(n):
    if not is_nondegenerate_nucleotide_string(n):
        n = n.upper().replace('U', 'T')
        return all([_ in DEGENERATE_NUCLEOTIDES for _ in n])
    return False


def is_valid_nucleotide_string(s):
    """
    Checks if string is composed of valid nucleotide members.

    input is first capitalized and U's are replaced with T's
    each letter must be a standard or degenerate letter
    """
    s = s.upper().replace('U', 'T')
    return all([_ in DEGENERATE_NUCLEOTIDES for _ in s])


def combine_nucleotides(*nucleotides, proportions=None):
    """
    Combines nucleotides, creating a new Nucleotide instance with
    a nucleotide composition as a weighted sum input compositions.

    Usage
    -----
    c = Nucleotide('C')
    t = Nucleotide('T')

    c_plus_t = combine_nucleotides(c, t)
    print(c_plus_t) --> "Nucleotide('N', " \
                        "composition={'A':0, 'C':0.5, 'G':0, 'T':0.5})"

""    c_plus_ttt = combine_nucleotides(c, t, proportions=[1, 3])
    print(c_plus_ttt) --> "Nucleotide('n', "\
                          "composition={'A':0, 'C':0.25, 'G':0, 'T':0.75})"
    """

    if (proportions is None) or (len(nucleotides) != len(proportions)):
        proportions = [1, ] * len(nucleotides)

    data = {}
    for n in STANDARD_NUCLEOTIDES:
        data[n] = 0
        for nuc, p in zip(nucleotides, proportions):
            data[n] += nuc.composition.get(n, 0) * p
    return Nucleotide(rescale(data))


def nucleotide_composition_to_letter(composition):
    """
    Converts dictionary of {nucleotide letter: proportions} pairs
    to IUPAC degenerate DNA letter.

    Usage:
    c = {'A': 1}
    print(nucleotide_composition_to_letter(c)) --> 'A'

    c = dict(zip('ACGT', [1, 1, 1, 1]))
    print(nucleotide_composition_to_letter(c)) --> 'N'

    c = dict(zip('ACGT', [1, 1, 2, 1]))
    print(nucleotide_composition_to_letter(c)) --> 'n'
    """

    nonzero_nucleotides = ''.join(sorted([n
                                          for n, v in composition.items()
                                          if v > 0]))
    nonzero_proportions = [composition[n] for n in nonzero_nucleotides]
    equimolar = min(nonzero_proportions) == max(nonzero_proportions)

    letter = DEGENERATE_NUCLEOTIDE_CODE_REVERSED.get(nonzero_nucleotides,
                                                     DEFAULT_NUCLEOTIDE_LABEL)

    if equimolar:
        return letter
    return letter.lower()



#!/usr/bin/env python
# ~*~# -*- coding: utf-8 -*-

"""
The genetic code defines the mapping between nucleic acids and amino acids.

This module defines the GeneticCode object and related functions.

Usage
*****
std = GeneticCode()  # standard genetic code
std['AAA'] -> 'K'
std['BAD_CODON'] -> '*' # stop symbol
"""


from dogma import (
    NBCI_GENETIC_CODE_NAMES,
    NCBI_CODONS,
    NCBI_AMINO_ACIDS
)


class GeneticCode:
    """
    System for translating codons into amino acids.

    Parameters
    ----------
    ncbi_id: integer witin NBCI_GENETIC_CODE_NAMES keys,
             default is 1, standard genetic code
    updated_mappings: dictionary
    stop_symbol: string with length 1
    error_symbol: string with length 1
    name: str
    """

    def __init__(self,
                 ncbi_id=1,
                 updated_mappings={},
                 stop_symbol='*',
                 error_symbol='_',
                 name=''):

        if ncbi_id in NBCI_GENETIC_CODE_NAMES:
            name = NBCI_GENETIC_CODE_NAMES[ncbi_id]
            codons = NCBI_CODONS
            amino_acids = NCBI_AMINO_ACIDS[ncbi_id]
        else:
            codons = []
            amino_acids = ''

        if updated_mappings:
            assert isinstance(updated_mappings, dict)
            assert all([(len(c) == 3 and isinstance(c, str))
                        for c in updated_mappings.keys()])
            assert all([(len(a) == 1 and isinstance(a, str))
                        for a in updated_mappings.values()])

        if error_symbol is not '_':
            assert isinstance(error_symbol, str) and len(error_symbol) == 1
        self.error_symbol = error_symbol

        if stop_symbol is not '*':
            assert isinstance(stop_symbol, str) and len(stop_symbol) == 1
            amino_acids = amino_acids.replace('*', stop_symbol)
            self.stop_symbol = stop_symbol

        if isinstance(name, str):
            self.name = name

        # define genetic code mapping from codons to amino_acids
        self.code = dict(zip(codons, amino_acids))

        # overwrite specified codon to amino acid mappings
        for codon, amino_acid in updated_mappings.items():
            self.code[codon] = amino_acid

        # create list of codons and amino acids
        self.codons = codons
        self.amino_acids = amino_acids

        # define reversed genetic code, where each amino acids
        # maps to a list of codons
        self.code_reversed = {a: [_ for _ in codons
                                  if self.code[_] == a]
                              for a in amino_acids}

    def update(self, updated_mappings={}, expand=False):
        """
        Updates genetic code, reprogramming codons to amino acids

        Parameters
        ----------
        updated_mappings: dict

        expand: Boolean
            if True, genetic code is 'expanded' to include new codons
            if False, only keys in GeneticCode.code.keys() will be remapped,
                      others silently ignored.
            Default is False

        Usage
        -----
        standard = GeneticCode()  #  default parameter ncbi_id = 1
                                  #  corresponds to 'Standard' code
        supE = GeneticCode(updated_mappings={'TAG': 'Q'})
        """
        assert isinstance(updated_mappings, dict)
        assert all([(len(c) == 3 and isinstance(c, str))
                    for c in updated_mappings.keys()])
        assert all([(len(a) == 1 and isinstance(a, str))
                    for a in updated_mappings.values()])

        if not expand:
            updated_mappings = {k: v
                                for k, v in updated_mappings.items()
                                if k in self.code}

        for codon, amino_acid in updated_mappings.items():
            self.code[codon] = amino_acid

    def get_codons(self):
        """
        Returns sorted list of codons in genetic code.
        """
        return sorted(list(self.code.keys()))

    def get_amino_acids(self):
        """
        Returns list of amino acids, ordered by sorted codons.
        """
        return [self.code[codon] for codon in self.get_codons()]

    def __getitem__(self, codon):
        """
        Returns value of code dictionary for specified codon key.

        Parameter
        ---------
        codon: string within code dictionary

        Returns
        -------
        One-letter amino acid string

        Usage
        -----
        standard = GeneticCode(1)
        standard['GGG'] --> 'G'
        """
        codon = codon.upper().replace('U', 'T')

        return self.code.get(codon, self.error_symbol)

    def __setitem__(self, codon, amino_acid):
        """
        Sets value of code dictionary for specified codon key
        to specified amino acid.

        Parameter
        ---------
        codon: string within genetic code dictionary

        Usage
        -----
        >> std = GeneticCode()
        >> std['GGG'] --> 'G'
        >> std['GGG'] = 'Z'
        >> std['GGG'] --> 'Z'

        """
        codon = codon.upper().replace('U', 'T')
        if codon not in self.code:
            raise "Attempting to set codon not in genetic code."
        else:
            self.code[codon] = amino_acid

    def __str__(self):
        """
        Returns readable genetic code list
        """
        codons = self.get_codons()
        amino_acids = self.get_amino_acids()
        return '\n'.join([' '.join([c, a])
                         for c, a in zip(codons, amino_acids)])

    def __repr__(self):
        """
        Returns string that would evaluate to an equivalent
        GeneticCode instance.
        """
        codons = self.get_codons()
        amino_acids = self.get_amino_acids()
        return f'GeneticCode(codons={codons}, amino_acids={amino_acids})'


DEFAULT_GENETIC_CODE = GeneticCode()

SUPE = GeneticCode(updated_mappings={'TAG': 'Q'}, name='supE')


def translate(dna, genetic_code=None):
    """
    Translates a DNA string into amino acid string.

    RNA and lower-case strings accepted.
    """
    assert isinstance(dna, str)

    # process string by replacing 'U's with 'T's and uppercasing
    dna = dna.upper().replace('U', 'T')

    if not isinstance(genetic_code, GeneticCode):
        genetic_code = DEFAULT_GENETIC_CODE

    codon_strings = [dna[3 * i: 3 * i + 3] for i in range(len(dna) // 3)]

    return ''.join([genetic_code[_] for _ in codon_strings])


def test():
    for k, v in SUPE.code.items():
        print(f"'{k}': '{v}',")


if __name__ == '__main__':
    test()
